# -*- coding: utf-8 -*-
"""Some of the graph-like connectivity algorithms which are typically applied to
meshes but may be used elsewhere are exposed here."""

from ._polygon_walk import connected, group_connected
