# -*- coding: utf-8 -*-
"""
"""

from ._geometry import (
    area,
    center_of_mass,
    closest,
    furthest,
    get_components,
    get_components_zipped,
    inner_product,
    magnitude,
    magnitude_sqr,
    normalise,
    normalised,
    orthogonal_bases,
    snap_to_plane,
    unzip,
    zip,
)
from ._geometry import _reduce_last_axis
from ._unit_vector import UnitVector
